# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

class Defaults:
    PORTAL_URL = 'https://portal.tol.sanger.ac.uk'
    PORTAL_API_PATH = '/api/v1'
    PORTAL_API_DATA_PATH = '/data'

    STS_URL = 'https://sts.tol.sanger.ac.uk'
    STS_API_PATH = '/api/v2'
    STS_API_DATA_PATH = '/data'

    TOLID_URL = 'https://id.tol.sanger.ac.uk'
    TOLID_API_PATH = '/api/v3'
    TOLID_API_DATA_PATH = ''

    TOLQC_URL = 'https://qc.tol.sanger.ac.uk'
    TOLQC_API_PATH = '/api/v1'
    TOLQC_API_DATA_PATH = '/data'

    TREEOFSEX_URL = 'https://treeofsex.sanger.ac.uk'
    TREEOFSEX_API_PATH = '/api/v1'
    TREEOFSEX_API_DATA_PATH = '/data'

    BIOSCAN_URL = 'https://bioscan.tol.sanger.ac.uk'
    BIOSCAN_API_PATH = '/api/v1'
    BIOSCAN_API_DATA_PATH = ''

    BIOSCAN_IMAGE_BUCKET_NAME = 'tol-bioscan-images'
    BIOSCAN_IMAGE_PREFIX = 'processed_images/'

    WORKFLOWS_URL = 'https://workflows.tol.sanger.ac.uk'
    WORKFLOWS_API_PATH = '/api/v2'
    WORKFLOWS_API_DATA_PATH = '/data'

    GOAT_URL = 'https://goat.genomehubs.org'
    GOAT_API_PATH = '/api/v2'

    BOLD_URL = 'https://data.boldsystems.org'
    BOLD_PORTAL_URL = 'https://portal.boldsystems.org'
    BOLD_API_PATH = '/api'

    COPO_URL = 'https://copo-project.org'
    COPO_API_PATH = '/api'

    JIRA_URL = 'https://jira.sanger.ac.uk'

    LABWHERE_URL = 'https://labwhere.psd.sanger.ac.uk'
    LABWHERE_API_PATH = '/api'

    ENA_URL = 'https://www.ebi.ac.uk'
